package cn.iocoder.yudao.module.member.mq.consumer.score.core;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.util.ArrayUtil;
import cn.hutool.json.JSONUtil;
import cn.iocoder.boot.module.order.api.OrderApi;
import cn.iocoder.boot.module.order.api.dto.OrderItemRespDTO;
import cn.iocoder.boot.module.order.api.dto.OrderOperateLogDTO;
import cn.iocoder.boot.module.order.api.dto.OrderRespDTO;
import cn.iocoder.boot.module.order.enums.OrderStatusApiEnum;
import cn.iocoder.yudao.module.customer.dal.dataobject.customerContacts.CustomerContactsDO;
import cn.iocoder.yudao.module.customer.service.customerContacts.CustomerContactsService;
import cn.iocoder.yudao.module.member.api.score.MemberUserScoreApi;
import cn.iocoder.yudao.module.member.api.score.dto.MemberUserScoreOperateReqDTO;
import cn.iocoder.yudao.module.member.dal.dataobject.scoreRule.ScoreRuleDO;
import cn.iocoder.yudao.module.member.enums.*;
import cn.iocoder.yudao.module.member.mq.message.ScoreMessage;
import cn.iocoder.yudao.module.member.service.scoreLog.MemberUserScoreLogService;
import cn.iocoder.yudao.module.member.service.scoreRule.ScoreRuleService;
import cn.iocoder.yudao.module.member.service.user.MemberUserService;
import cn.iocoder.yudao.module.member.vo.scoreRule.extra.ScoreRuleOrderVExtraVO;
import com.alibaba.fastjson.JSONObject;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.redisson.RedissonMultiLock;
import org.redisson.api.RLock;
import org.redisson.api.RedissonClient;
import org.springframework.stereotype.Service;

import javax.annotation.Resource;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

import static cn.iocoder.yudao.module.member.enums.TransportTypeEnum.SPECIAL_LINE_AIR_FREIGHT;

/**
 * @author zhangfeng
 */
@Service
@Slf4j
public class OrderVStrategy extends AbstractScoreRuleStrategy {
    @Resource
    private OrderApi orderApi;
    @Resource
    private CustomerContactsService customerContactsService;
    @Resource
    private RedissonClient redissonClient;

    public OrderVStrategy(ScoreRuleService scoreRuleService, MemberUserScoreLogService memberUserScoreLogService, MemberUserService memberUserService, MemberUserScoreApi memberUserScoreApi) {
        super(scoreRuleService, memberUserScoreLogService, memberUserService, memberUserScoreApi);
    }

    @Override
    public void addScore(ScoreMessage message) {
        Long orderId = message.getOrderId();
        try {
            Thread.sleep(3000);
        } catch (InterruptedException e) {
            log.error("Order in shipping listener: Thread sleep error", e);
        }
        // 获取订单信息
        OrderRespDTO order;
        try {
            order = orderApi.getOrder(orderId, message.getOrderNo());
        } catch (Exception e) {
            log.error("Order in shipping listener: Get order error,orderId:{}", orderId == null ? message.getOrderNo() : orderId, e);
            return;
        }
        if (Objects.isNull(order)) {
            log.error("Order in shipping listener: The order does not exist,orderId:{}", orderId == null ? message.getOrderNo() : orderId);
            return;
        }
        orderId = order.getOrderId();
        log.info("Received order V message,orderID:{}", orderId);
        OrderOperateLogDTO orderOperateLogDTO = new OrderOperateLogDTO();
        orderOperateLogDTO.setOrderId(orderId);
        orderOperateLogDTO.setCreatorName("订单V值规则");
        orderOperateLogDTO.setType("积分操作");
        orderOperateLogDTO.setUserType(2);
        orderOperateLogDTO.setMsg("订单起飞/起运计算积分");
        StringBuilder logBody = new StringBuilder();
        // 订单状态应为起运
        if (!Objects.equals(order.getStatus(), OrderStatusApiEnum.IN_SHIPPING.getValue())) {
            log.info("Order in shipping listener: The order status is not in shipping,orderID:{}", orderId);
            return;
        }
        // 增加订单无收货人时校验
        long consigneeId = 0;
        if (Objects.isNull(order.getOrderConsigneeDO()) || Objects.isNull(order.getOrderConsigneeDO().getCustomerId())) {
            log.info("Order in shipping listener: The order consignee is empty,orderID:{}", orderId);
        } else {
            consigneeId = order.getOrderConsigneeDO().getCustomerId();
        }
        // 获取订单规则
        List<ScoreRuleDO> scoreRuleDOList = this.getMatchRule(order, consigneeId);
        if (Objects.isNull(scoreRuleDOList)) {
            log.info("Order in shipping listener: No score rule match,orderID:{}", orderId);
            orderOperateLogDTO.setBody(JSONObject.toJSONString(logBody.append("无匹配积分规则；")));
            orderApi.createOrderOperateLog(orderOperateLogDTO);
            return;
        }
        for (int i = 0; i < scoreRuleDOList.size(); i++) {
            ScoreRuleDO scoreRuleDO = scoreRuleDOList.get(i);
            orderOperateLogDTO.setCreator(scoreRuleDO.getCode());
            ScoreRuleOrderVExtraVO extraInfo = JSONUtil.toBean(scoreRuleDO.getExtra(), ScoreRuleOrderVExtraVO.class);
            String lockKey1;
            if (order.getUserId() > 0) {
                lockKey1 = "score:rule:order:lock:" + order.getUserId();
            } else {
                lockKey1 = "score:rule:order:lock:" + order.getOrderConsignorDO().getCustomerId();
            }
            String lockKey2 = "score:rule:order:lock:" + consigneeId;
            RLock lock1 = redissonClient.getLock(lockKey1);
            RLock lock2 = consigneeId == 0 ? null : redissonClient.getLock(lockKey2);
            RedissonMultiLock multiLock = new RedissonMultiLock(lock1, lock2);
            try {
                boolean tryLock = multiLock.tryLock(2, 10, TimeUnit.SECONDS);
                if (!tryLock) {
                    log.error("Order in shipping listener: get lock error, orderId:{}", orderId);
                }
                // 获取需要添加积分的会员
                List<Long> userIds = getAddScoreUser(order.getUserId(), order.getOrderConsignorDO().getCustomerId(), consigneeId, scoreRuleDO, logBody);
                if (CollectionUtil.isEmpty(userIds)) {
                    log.info("Order in shipping listener: consignor and consignee are not bound to the member, orderId:{}", orderId);
                    orderOperateLogDTO.setBody(JSONObject.toJSONString(logBody));
                    orderApi.createOrderOperateLog(orderOperateLogDTO);
                    continue;
                }
                // 计算积分
                Integer scoreCount = getScoreCount(order.getTransportId(), extraInfo.getOrderVRule(), orderId);
                logBody.append("根据").append(scoreRuleDO.getCode()).append("规则计算积分为：").append(scoreCount).append("；");
                if (scoreCount == 0) {
                    log.info("Order in shipping listener: The score count is less than or equal to 0,orderId:{}", orderId);
                    orderOperateLogDTO.setBody(JSONObject.toJSONString(logBody));
                    orderApi.createOrderOperateLog(orderOperateLogDTO);
                    continue;
                }
                // 增加积分
                for (Long userId : userIds) {
                    try {
                        HashMap<String, Object> map = new HashMap<>(1);
                        map.put("orderId", orderId);
                        memberUserScoreApi.operateScore(MemberUserScoreOperateReqDTO.builder()
                                .memberId(userId)
                                .scoreCount(scoreCount)
                                .sourceType(ScoreSourceTypeEnum.ORDER_V)
                                .ruleId(scoreRuleDO.getId())
                                .expireDays(scoreRuleDO.getScorePeriod())
                                .releationId(String.valueOf(orderId))
                                .uniqueId(orderId + "_" + userId)
                                .extParam(map)
                                .build()
                        );
                        log.info("Order V rule add score success,ruleID:{},orderID:{},userID:{}", scoreRuleDO.getId(), orderId, userId);
                    } catch (Exception e) {
                        log.error("Order in shipping listener: operateScore error", e);
                    }
                }
            } catch (InterruptedException e) {
                log.error("Order in shipping listener: lock error orderId:{}", orderId, e);
            } finally {
                multiLock.unlock();
            }
        }

        orderOperateLogDTO.setBody(JSONObject.toJSONString(logBody.append("获取积分成功")));
        orderApi.createOrderOperateLog(orderOperateLogDTO);
    }

    /**
     * 根据规则获得要加多少分
     *
     * @param transportId
     * @param orderVRules
     * @param orderId
     * @return
     */
    private Integer getScoreCount(Integer transportId, List<ScoreRuleOrderVExtraVO.OrderVRule> orderVRules, Long orderId) {
        Integer scoreCount = 0;
        List<OrderItemRespDTO> orderItems = orderApi.getOrderItemsByOrderId(orderId);
        if (orderItems == null) {
            log.info("Order in shipping listener: The order item is null,orderId:{}", orderId);
            return 0;
        }
        BigDecimal totalWeight = BigDecimal.ZERO;
        BigDecimal totalVolume = BigDecimal.ZERO;
        for (OrderItemRespDTO orderItem : orderItems) {
            totalWeight = totalWeight.add(orderItem.getChargeWeight());
            totalVolume = totalVolume.add(orderItem.getChargeVolume());
        }
        if (transportId == TransportTypeEnum.OCEAN_LCL.getValue()) {
            // 海运算体积
            scoreCount = computeScoreCount(totalVolume, orderVRules);
        } else if (transportId == SPECIAL_LINE_AIR_FREIGHT.getValue()) {
            // 空运算重量
            scoreCount = computeScoreCount(totalWeight, orderVRules);
        }
        return scoreCount;
    }

    /**
     * 获取需要加分的用户
     */
    private List<Long> getAddScoreUser(Long userId, Long consignorId, Long consigneeId, ScoreRuleDO scoreRuleDO, StringBuilder logNote) {
        ArrayList<Long> users = new ArrayList<>();
        ScoreRuleOrderVExtraVO extraInfo = JSONUtil.toBean(scoreRuleDO.getExtra(), ScoreRuleOrderVExtraVO.class);
        String customerSide = extraInfo.getCustomerSide();
        String[] customerSideArray = customerSide.split(",");
        //校验客户方是否存在且是否有收货人
        boolean hasConsignee =ArrayUtil.contains(customerSideArray, String.valueOf(CustomerSideEnum.CONSIGNEE.getValue()));
        //校验客户方是否存在且是否有发货人
        boolean hasConsignor =ArrayUtil.contains(customerSideArray, String.valueOf(CustomerSideEnum.CONSIGNOR.getValue()));
        // 发货人和收货人都要判断给分
        boolean isBackendOrder = userId <= 0;
        logNote.append("匹配规则"+scoreRuleDO.getCode()+"   此订单为").append(isBackendOrder ? "后台" : "客户端").append("下单；");
        // 如果是后台下单，重新获取发货人用户ID
        if (isBackendOrder) {
            CustomerContactsDO contactsDO = getCustomerMemberId(consignorId);
            if (contactsDO == null || contactsDO.getUserid() == 0) {
                userId = 0L;
                logNote.append("发货人未绑定会员；");
            } else {
                if (!hasConsignor&&customerSideArray.length!=0){
                    logNote.append("发货人绑定会员：").append(contactsDO.getUsername()).append("，会员ID：").append(userId).append("；");
                    userId = 0L;
                    logNote.append("积分规则客户方不包含发货人   ");
                }
                if (hasConsignor||customerSideArray.length==0){
                    userId = contactsDO.getUserid();
                    logNote.append("发货人绑定会员：").append(contactsDO.getUsername()).append("，会员ID：").append(userId).append("；");
                }
            }
        }else {
            CustomerContactsDO contactsDO = getCustomerMemberId(consignorId);
            if (!hasConsignor&&customerSideArray.length!=0){
                logNote.append("发货人绑定会员：").append(contactsDO.getUsername()).append("，会员ID：").append(consignorId).append("；");
                userId = 0L;
                logNote.append("积分规则客户方不包含发货人");
            }
            if (hasConsignor||customerSideArray.length==0){
                logNote.append("发货人绑定会员：").append(contactsDO.getUsername()).append("，会员ID：").append(consignorId).append("；");
            }
        }
        // 校验积分上限，添加加分用户列表
        if (isBackendOrder && isUserScoreMaxed(userId, scoreRuleDO)) {
            log.info("Order in shipping listener: The user has reached the maximum score, userId:{}, scoreRuleId:{}", userId, scoreRuleDO.getId());
            logNote.append("发货人").append(scoreRuleDO.getCode()).append("规则积分已到上限；");
        } else {
            addValidUser(users, userId);
        }

        // 判断收货人
        if (consigneeId == null || consigneeId <= 0) {
            logNote.append("此订单无收货人；");
            return users;
        }
        CustomerContactsDO consigneeMemberDO = getCustomerMemberId(consigneeId);
        if (consigneeMemberDO == null) {
            logNote.append("收货人未绑定会员；");
        }
        Long consigneeMemberId = consigneeMemberDO == null ? 0L : consigneeMemberDO.getUserid();
        if (consigneeMemberId > 0) {
            logNote.append("收货人绑定会员：").append(consigneeMemberDO.getUsername()).append("，会员ID：").append(consigneeMemberId).append("；");
            if (consigneeMemberId.equals(userId)) {
                log.info("Order in shipping listener: The consignee is the same as the sender, userId:{}", userId);
            } else if (isBackendOrder && isUserScoreMaxed(consigneeMemberId, scoreRuleDO)) {
                logNote.append("收货人").append(scoreRuleDO.getCode()).append("规则积分已到上限；");
                log.info("Order in shipping listener: The consignee has reached the maximum score, userId:{}, scoreRuleId:{}", consigneeMemberId, scoreRuleDO.getId());
            } else {
                if (!hasConsignee&&customerSideArray.length!=0){
                    logNote.append("积分规则客户方不包含收货人   ");
                }
                if (hasConsignee||customerSideArray.length==0){
                    users.add(consigneeMemberId);
                }
            }
        }
        return users;
    }

    /**
     * 校验订单入口是否有效
     */
    private boolean isOrderEntryValid(String orderEntryString, boolean isBackendOrder) {
        if (StringUtils.isBlank(orderEntryString)) {
            log.info("Order in shipping listener: orderEntry is null");
            return false;
        }
        String[] orderEntry = orderEntryString.split(",");
        if (isBackendOrder) {
            return ArrayUtil.contains(orderEntry, String.valueOf(PlatformTypeEnum.BACKEND.getValue()));
        } else {
            return ArrayUtil.contains(orderEntry, String.valueOf(PlatformTypeEnum.APP.getValue())) ||
                    ArrayUtil.contains(orderEntry, String.valueOf(PlatformTypeEnum.WEB.getValue()));
        }
    }

    /**
     * 获取匹配的规则
     *
     * @param order
     * @param consigneeId
     * @return
     */
    private List<ScoreRuleDO> getMatchRule(OrderRespDTO order, long consigneeId) {
        List<ScoreRuleDO> scoreRuleDOs = scoreRuleService.getEnableOrderVRule();
        if (CollectionUtil.isEmpty(scoreRuleDOs)) {
            return null;
        }
        // 运输方式
        Integer transportId = order.getTransportId();
        // 首单
        Integer isFirstOrder = isFirstOrderValid(order.getOrderConsignorDO().getCustomerId(), consigneeId);
        // 提货点
        String warehouseId = order.getOrderObjectiveDO().getObjectiveWarehouseId().toString();
        // 渠道
        Long channelId = order.getChannelId();
        // 订单入口
        boolean isBackendOrder = order.getUserId() <= 0;

        return scoreRuleDOs.stream().filter(scoreRuleDO -> {
            ScoreRuleOrderVExtraVO extraInfo = JSONUtil.toBean(scoreRuleDO.getExtra(), ScoreRuleOrderVExtraVO.class);
            if (!isOrderEntryValid(extraInfo.getOrderEntry(), isBackendOrder)) {
                return false;
            }
            if (!transportId.equals(extraInfo.getTransportType())) {
                return false;
            }
            if (extraInfo.getFirstOrder()!=2&&!isFirstOrder.equals(extraInfo.getFirstOrder())) {
                return false;
            }
            String[] warehouseIds = extraInfo.getReceiveAddr().split(",");
            // 0代表所有提货点
            if (!Objects.equals(warehouseIds[0], "0") && !ArrayUtil.contains(warehouseIds, warehouseId)) {
                return false;
            }
            if (transportId.equals(SPECIAL_LINE_AIR_FREIGHT.getValue())) {
                if (!extraInfo.getChannel().contains(channelId)) {
                    return false;
                }
            }
            return true;
        }).collect(Collectors.toList());
    }

    /**
     * 校验是否为首单
     */
    private Integer isFirstOrderValid(Long consignorId, Long consigneeId) {
        // 发货和收货都算首单，按客户查，只有发货人收货人都是首单才算首单
        if (!orderApi.isFirstOrder(consignorId)) {
            log.info("Order in shipping listener: consignor not first order,consignorId:{}", consignorId);
            return 1;
        }
        if (consigneeId != 0 && !orderApi.isFirstOrder(consigneeId)) {
            log.info("Order in shipping listener: consignee not first order,consigneeId:{}", consigneeId);
            return 1;
        }
        return 0;
    }

    /**
     * 校验用户积分是否达到上限
     */
    private boolean isUserScoreMaxed(Long userId, ScoreRuleDO scoreRuleDO) {
        if (scoreRuleDO.getMaxScoreTotal() > 0) {
            int userScoreTotalCount = getUserScoreTotalCount(scoreRuleDO.getId(), userId);
            return userScoreTotalCount >= scoreRuleDO.getMaxScoreTotal();
        }
        return false;
    }

    /**
     * 添加符合条件的用户到列表
     */
    private void addValidUser(List<Long> users, Long userId) {
        if (userId != null && userId > 0) {
            users.add(userId);
        }
    }

    /**
     * 根据客户id获取会员
     *
     * @param customerId
     * @return
     */
    private CustomerContactsDO getCustomerMemberId(Long customerId) {
        List<CustomerContactsDO> customerContacts = customerContactsService.getCustomerContactsListByCustomerId(customerId);
        if (!CollectionUtil.isEmpty(customerContacts)) {
            for (CustomerContactsDO customerContact : customerContacts) {
                if (customerContact.getIsDefault() == 1 && customerContact.getUserid() != null) {
                    return customerContact;
                }
            }
        }
        log.info("Order in shipping listener: The customer is not bound to the member or have default userId, customerId:{}", customerId);
        return null;
    }

    @Override
    public ScoreRuleTypeEnum getStrategyScoreRuleType() {
        return ScoreRuleTypeEnum.ORDER_V;
    }


    /**
     * 根据规则计算获取积分
     *
     * @param weightOrVolume 重量或体积
     * @param orderVRules    规则
     * @return
     */
    private Integer computeScoreCount(BigDecimal weightOrVolume, List<ScoreRuleOrderVExtraVO.OrderVRule> orderVRules) {
        if (orderVRules.isEmpty()) {
            return 0;
        }
        int wOrV = weightOrVolume.setScale(0, RoundingMode.HALF_UP).intValue();
        ScoreRuleOrderVExtraVO.OrderVRule lastRule = orderVRules.get(orderVRules.size() - 1);
        if (wOrV >= lastRule.getHigh()) {
            return lastRule.getScore() * wOrV;
        }
        for (ScoreRuleOrderVExtraVO.OrderVRule rule : orderVRules) {
            if (rule.getLow() <= wOrV && wOrV < rule.getHigh()) {
                return rule.getScore() * wOrV;
            }
        }
        return 0;
    }
}
